import click
from .connection import Connection


@click.group('connection')
@click.pass_context
def connection(ctx):
    """Connection related subcommands"""
    ctx.obj = Connection()


@click.command('down')
@click.argument('name', type=str, required=True)
@click.pass_obj
def down(conn, name):
    """Deactivate a connection"""
    conn.deactivate(name)


@click.command('list')
@click.pass_obj
def _list(conn):
    """List connections"""
    conn.show_all()


@click.command('active')
@click.pass_obj
def list_active(conn):
    """List active connections"""
    conn.show_active()


@click.command('delete')
@click.argument('name', type=str, required=True)
@click.pass_obj
def delete(conn, name):
    """Delete a connection profile"""
    conn.delete(name)


connection.add_command(down)
connection.add_command(_list)
connection.add_command(list_active)
connection.add_command(delete)



import pytest
from unittest.mock import call
from wypy.utils.constants import NM_BUS_NAME, NM_DEVICE_IFACE

fmt_mock_path = 'wypy.device.device.format_table_key'
colored_mock_path = 'wypy.device.device.colored'


def test_fill_details_table(device, mocker):

    data = {
        'name': 'karim',
        'location': 'paris',
        'player': 'benzema',
        'birth_place': 'lyon'
    }
    table = device.details_table
    fmt_mock = mocker.patch(fmt_mock_path)
    colored_mock = mocker.patch(colored_mock_path)
    add_row_mock = mocker.patch.object(table, 'add_row')

    device._fill_details_table(data)

    fmt_mock_call_args = [call(k) for k in data.keys()]

    fmt_mock.assert_has_calls(fmt_mock_call_args, any_order=True)
    assert len(colored_mock.call_args_list) == len(data.values())
    assert len(add_row_mock.call_args_list) == len(data.values())


def test_get_device_details(device, mocker):
    all_props_mock = mocker.patch.object(device, 'get_all_properties')
    str_dbus_vals_mock = mocker.patch.object(device, '_stringify_dbus_values')
    dev_status_mock = mocker.patch.object(device, '_get_device_status')

    result = device._get_device_details('some_obj_path')

    assert all_props_mock.called_once()
    assert str_dbus_vals_mock.called_once()
    assert dev_status_mock.called_once()

    assert isinstance(result, dict)
    assert 'mtu' in result.keys()


def test_get_device_details_show_all(device, mocker):
    all_props_mock = mocker.patch.object(device, 'get_all_properties')
    str_dbus_vals_mock = mocker.patch.object(device, '_stringify_dbus_values')
    dev_status_mock = mocker.patch.object(device, '_get_device_status')
    expected_keys = ['mtu', 'ipv4_addresses', 'ipv4_domains', 'ipv4_dns', 'ipv4_gateway']

    result = device._get_device_details('some_obj_path', show_all=True)

    assert all_props_mock.called_once()
    assert str_dbus_vals_mock.called_once()
    assert dev_status_mock.called_once()

    assert isinstance(result, dict)
    assert sorted(expected_keys) == sorted(result.keys())


def test_get_connection_name(device, mocker):
    conn_name = 'my_conn_name'
    all_props_mock = mocker.patch.object(
        device,
        'get_all_properties',
        return_value={'Id': conn_name}
    )
    result = device._get_connection_name('some_connection_path')

    all_props_mock.assert_called_once()
    assert result == conn_name


def test_get_connection_name_exc(device, mocker):
    echo_mock = mocker.patch('click.echo')
    with pytest.raises(Exception):
        device._get_connection_name('some_conn_path')
    echo_mock.assert_called_once()


def test_create_row(device, mocker):
    row_color_mock = mocker.patch.object(
        device, 
        'get_device_state_row_color',
        return_value="green"
    )

    state = 10

    data = {
        'name': 'karim',
        'location': 'paris',
        'player': 'benzema',
        'birth_place': 'lyon',
        'device_path': 'hello',
        'connection_path': 'hello',
        'state': state
    }

    initial_values = len(data.values())

    result = device._create_row(data)

    row_color_mock.assert_called_once_with(10)

    assert len(result) == initial_values - 3


def test_get_device_status(device, mocker):
    all_props_mock = mocker.patch.object(device, 'get_all_properties')
    str_dbus_vals_mock = mocker.patch.object(device, '_stringify_dbus_values')
    conn_name_mock = mocker.patch.object(device, '_get_connection_name')
    translate_dev_state = mocker.patch.object(device, 'translate_device_state')
    translate_dev_type = mocker.patch.object(device, 'translate_device_type')

    expected_keys = [
        'name', 'type',
        'device_status', 'connection',
        'state', 'connection_path', 'device_path'
    ]
    result = device._get_device_status('some_object_path')

    all_props_mock.assert_called_once()
    str_dbus_vals_mock.assert_called_once()
    conn_name_mock.assert_called_once()
    translate_dev_type.assert_called_once()
    translate_dev_state.assert_called_once()

    assert sorted(result.keys()) == sorted(expected_keys)


def test_print_status(device, mocker):
    echo_mock = mocker.patch('click.echo')
    add_row_mock = mocker.patch.object(device.status_table, 'add_row')
    dev_status_mock = mocker.patch.object(device, '_get_device_status')
    create_row_mock = mocker.patch.object(
        device,
        '_create_row',
        return_value=['hello', 'world', 'liverpool', 'mancity']
    )
    device.all_devices = ['/dev/1', '/dev/2', '/dev/3']

    device.print_status()

    assert len(echo_mock.call_args_list) == 2
    assert len(add_row_mock.call_args_list) == len(device.all_devices)
    assert len(create_row_mock.call_args_list) == len(device.all_devices)
    assert len(dev_status_mock.call_args_list) == len(device.all_devices)


def test_list_all(device, mocker):
    dev_details_mock = mocker.patch.object(device, '_get_device_details')
    fill_details_mock = mocker.patch.object(device, '_fill_details_table')
    echo_mock = mocker.patch('click.echo')
    clear_rows_mock = mocker.patch.object(device.details_table, 'clear_rows')

    device.all_devices = ['/dev/1', '/dev/2', '/dev/3']

    device.list_all()

    assert len(echo_mock.call_args_list) == len(device.all_devices) + 1
    assert len(clear_rows_mock.call_args_list) == len(device.all_devices)
    assert len(dev_details_mock.call_args_list) == len(device.all_devices)
    assert len(fill_details_mock.call_args_list) == len(device.all_devices)


def test_manage_device_on(device, mocker):
    dummy_path = '/some/path/to/device'
    dummy_iface = 'dummy0'
    dummy_obj = 'some_obj'
    dev_path_mock = mocker.patch.object(
        device,
        '_get_known_device_object_path',
        return_value=dummy_path
    )
    get_obj_mock = mocker.patch.object(
        device.bus,
        'get_object',
        return_value=dummy_obj

    )
    set_prop_mock = mocker.patch.object(device, 'set_object_property')

    device.manage(dummy_iface)

    dev_path_mock.assert_called_once_with(dummy_iface)
    get_obj_mock.assert_called_once_with(NM_BUS_NAME, dummy_path)
    set_prop_mock.assert_called_once_with(
        dummy_obj,
        bus_name=NM_DEVICE_IFACE,
        prop_name='Managed',
        value=True
    )


def test_manage_device_off(device, mocker):
    dummy_path = '/some/path/to/device'
    dummy_iface = 'dummy0'
    dummy_obj = 'some_obj'
    dev_path_mock = mocker.patch.object(
        device,
        '_get_known_device_object_path',
        return_value=dummy_path
    )
    get_obj_mock = mocker.patch.object(
        device.bus,
        'get_object',
        return_value=dummy_obj

    )
    set_prop_mock = mocker.patch.object(device, 'set_object_property')

    device.manage(dummy_iface, flag=False)

    dev_path_mock.assert_called_once_with(dummy_iface)
    get_obj_mock.assert_called_once_with(NM_BUS_NAME, dummy_path)
    set_prop_mock.assert_called_once_with(
        dummy_obj,
        bus_name=NM_DEVICE_IFACE,
        prop_name='Managed',
        value=False
    )


def test_manage_device_unknown(device, mocker):
    device.known_device_names = ['/dev/1', '/dev/2', '/dev/3']
    dummy_iface = 'unknown_iface'

    with pytest.raises(SystemExit) as exc:
        device.manage(dummy_iface)

    assert dummy_iface in str(exc.value)


def test_autoconnect_device_on(device, mocker):
    dummy_path = '/some/path/to/device'
    dummy_iface = 'dummy0'
    dummy_obj = 'some_obj'
    dev_path_mock = mocker.patch.object(
        device,
        '_get_known_device_object_path',
        return_value=dummy_path
    )
    get_obj_mock = mocker.patch.object(
        device.bus,
        'get_object',
        return_value=dummy_obj

    )
    set_prop_mock = mocker.patch.object(device, 'set_object_property')

    device.autoconnect(dummy_iface)

    dev_path_mock.assert_called_once_with(dummy_iface)
    get_obj_mock.assert_called_once_with(NM_BUS_NAME, dummy_path)
    set_prop_mock.assert_called_once_with(
        dummy_obj,
        bus_name=NM_DEVICE_IFACE,
        prop_name='Autoconnect',
        value=True
    )


def test_autoconnect_device_off(device, mocker):
    dummy_path = '/some/path/to/device'
    dummy_iface = 'dummy0'
    dummy_obj = 'some_obj'
    dev_path_mock = mocker.patch.object(
        device,
        '_get_known_device_object_path',
        return_value=dummy_path
    )
    get_obj_mock = mocker.patch.object(
        device.bus,
        'get_object',
        return_value=dummy_obj

    )
    set_prop_mock = mocker.patch.object(device, 'set_object_property')

    device.autoconnect(dummy_iface, flag=False)

    dev_path_mock.assert_called_once_with(dummy_iface)
    get_obj_mock.assert_called_once_with(NM_BUS_NAME, dummy_path)
    set_prop_mock.assert_called_once_with(
        dummy_obj,
        bus_name=NM_DEVICE_IFACE,
        prop_name='Autoconnect',
        value=False
    )


def test_autoconnect_device_unknown(device, mocker):
    device.known_device_names = ['/dev/1', '/dev/2', '/dev/3']
    dummy_iface = 'unknown_iface'

    with pytest.raises(SystemExit) as exc:
        device.autoconnect(dummy_iface)

    assert dummy_iface in str(exc.value)


def test_disconnect_device(device, mocker):
    dummy_iface = 'some_iface'
    dummy_path = '/some/path/to/device'
    dev_path_mock = mocker.patch.object(
        device,
        '_get_known_device_object_path',
        return_value=dummy_path
    )
    disconnect_mock = mocker.patch.object(device, '_disconnect_device')

    device.disconnect(dummy_iface)

    dev_path_mock.assert_called_once_with(dummy_iface)
    disconnect_mock.assert_called_once_with(dummy_path, dummy_iface)


def test_disconnect_device_unknown(device, mocker):
    device.known_device_names = ['/dev/1', '/dev/2', '/dev/3']
    dummy_iface = 'unknown_iface'

    with pytest.raises(SystemExit) as exc:
        device.disconnect(dummy_iface)

    assert dummy_iface in str(exc.value)


def test_update_device(device, mocker):
    dummy_iface = 'some_iface'
    dummy_path = '/some/path/to/device'
    dev_path_mock = mocker.patch.object(
        device,
        '_get_known_device_object_path',
        return_value=dummy_path
    )

    update_mock = mocker.patch.object(device, '_update_connection')

    device.update_ifname_connection(dummy_iface)

    dev_path_mock.assert_called_once_with(dummy_iface)
    update_mock.assert_called_once_with(dummy_path, dummy_iface)


def test_update_device_unknown(device, mocker):
    device.known_device_names = ['/dev/1', '/dev/2', '/dev/3']
    dummy_iface = 'unknown_iface'

    with pytest.raises(SystemExit) as exc:
        device.update_ifname_connection(dummy_iface)

    assert dummy_iface in str(exc.value)


def test_delete_device(device, mocker):
    dummy_iface = 'some_iface'
    dummy_path = '/some/path/to/device'
    dev_path_mock = mocker.patch.object(
        device,
        '_get_known_device_object_path',
        return_value=dummy_path
    )

    delete_mock = mocker.patch.object(device, '_delete_iface')

    device.delete_iface(dummy_iface)

    dev_path_mock.assert_called_once_with(dummy_iface)
    delete_mock.assert_called_once_with(dummy_path, dummy_iface)


def test_delete_device_unknown(device, mocker):
    device.known_device_names = ['/dev/1', '/dev/2', '/dev/3']
    dummy_iface = 'unknown_iface'

    with pytest.raises(SystemExit) as exc:
        device.delete_iface(dummy_iface)

    assert dummy_iface in str(exc.value)


# def test_disconnect_device_already_down(device, mocker):
#     """need dbus for that one"""
#     pass



VERSION = '1.0.0'
NM_BUS_NAME = 'org.freedesktop.NetworkManager'

NM_OBJ_PATH = '/org/freedesktop/NetworkManager'

NM_SETTINGS_OBJ_PATH = '/org/freedesktop/NetworkManager/Settings'
NM_SETTINGS_IFACE = 'org.freedesktop.NetworkManager.Settings'

NM_IFACE = 'org.freedesktop.NetworkManager'

NM_DEVICE_IFACE = 'org.freedesktop.NetworkManager.Device'

NM_ACTIVE_CONN_IFACE = 'org.freedesktop.NetworkManager.Connection.Active'

NM_CONNECTION_IFACE = 'org.freedesktop.NetworkManager.Settings.Connection'

IP4_CONFIG_IFACE = 'org.freedesktop.NetworkManager.IP4Config'

NM_WIRELESS_IFACE = 'org.freedesktop.NetworkManager.Device.Wireless'

NM_ACCESS_POINT_IFACE = 'org.freedesktop.NetworkManager.AccessPoint'

NM_CONNECTION_STATE_ACTIVATED = 2

DBUS_GENERAL_PROPS = {
    'Connectivity': 'CONNECTIVITY',
    'State': 'STATE',
    'WirelessEnabled': 'WIFI',
    'WwanEnabled': 'WWAN',
    'WirelessHardwareEnabled': 'WIFI-HW',
    'WwanHardwareEnabled': 'WWAN-HW'
}



from .wifi import WiFi
import click


@click.group('wifi')
@click.pass_context
def wifi(ctx):
    """Interact with the wireless device(s)"""
    ctx.obj = WiFi()


@click.command('on')
@click.pass_obj
def turn_wifi_on(wifi):
    """Enable WiFi"""
    wifi.turn_on()


@click.command('off')
@click.pass_obj
def turn_wifi_off(wifi):
    """Disable WiFi"""
    wifi.turn_off()


@click.command('status')
@click.pass_obj
def wifi_status(wifi):
    """Print current WiFi status"""
    wifi.print_status()


@click.command('list')
@click.pass_obj
def list_access_points(wifi):
    """List currently available access points"""
    wifi.list_access_points()


@click.command('rescan')
@click.pass_obj
def rescan(wifi):
    """Force WyPy to scan for available access points"""
    wifi.rescan()


@click.command('connect')
@click.pass_obj
def connect(wifi):
    """Connect to a wireless access point"""
    wifi.connect()


wifi.add_command(turn_wifi_on)
wifi.add_command(turn_wifi_off)
wifi.add_command(wifi_status)
wifi.add_command(list_access_points)
wifi.add_command(rescan)
wifi.add_command(connect)



from collections import MutableMapping
from uuid import UUID


def flatten(d, parent_key='', sep='.'):
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def nm_is_installed():
    """
    Checks if NetworkManager is intalled on the user's system

    Returns:
        bool -- if NetworkManager is installed
    """
    from shutil import which
    return which('NetworkManager') is not None


def is_valid_uuid(_str):
    """
    Checks if _str is a uuidv4 string
    return <bool> - Whether the string is a uuid
    """
    try:
        UUID(_str, version=4)
    except ValueError:
        return False
    else:
        return True


def format_list(data_list, key='address'):
    """
    this function will map over a given list of dictionaries,
    extract all instances dict[key] and join the
    result into a final string.

    It's particularly useful when joining lists of ip addresses
    and domain names.

    If the case `data_list` is empty, the function will simply return: '--'.
    """

    if len(data_list) == 0:
        return '--'
    # TODO: wrap this call in a try / except block
    ips = list(map(lambda x: str(x[key]), data_list))
    return ' - '.join(ips)


def format_table_key(key):
    """
    Accepts a string as the only argument
    replaces underscores with spaces and uppercases the string
    """
    return key.upper().replace('_', ' ')


def format_connection_name(name):
    """
        Takes a connection name and replaces
        every space with an underscore.
    """
    return name.replace(' ', '_')


def user_choice_to_bool(choice):
    """

    Arguments:
        choice {[string]} -- [the choice the user was prompted to make]

    Returns:
        [boolean] -- [the choice]
    """
    return True if choice == 'yes' else False



import pytest
import dbus
from termcolor import colored
from unittest.mock import call
from wypy.utils.constants import DBUS_GENERAL_PROPS, NM_DEVICE_IFACE


def test_turn_on_wifi(wifi, mocker):
    """
        Assert click.echo was called with the correct message passed
        as the argument.
        Assert WiFi._get_wifi_status_code was called and that if
        it returns 0 (False), Wifi._enable_wifi() is called
    """
    intro_msg = 'Enabling WiFi ...'

    echo_mock = mocker.patch('click.echo')
    get_status_mock = mocker.patch.object(wifi, '_get_wifi_status_code', return_value=0)
    wifi_mock = mocker.patch.object(wifi, '_enable_wifi')

    wifi.turn_on()

    echo_mock.assert_called_once_with(intro_msg)
    get_status_mock.assert_called_once()
    wifi_mock.assert_called_once()


def test_turn_on_wifi_error(wifi, mocker):
    """
        Assert click.echo was called with the correct message passed
        as the argument.
        Assert WiFi._get_wifi_status_code was called and that if
        it returns 1 (True), click.echo is called with the correct
        error message.
    """

    get_status_mock = mocker.patch.object(wifi, '_get_wifi_status_code', return_value=1)
    echo_mock = mocker.patch('click.echo')

    wifi.turn_on()

    expected_err_msg = call('WiFi is already enabled. Skipping.')
    get_status_mock.assert_called_once()
    echo_mock.assert_has_calls([expected_err_msg])


def test_turn_off_wifi(wifi, mocker):
    """
        Assert click.echo was called with the correct message passed
        as the argument.
        Assert WiFi._get_wifi_status_code was called and that if
        it returns 1 (enabled), WiFi._disable_wifi is called.
    """

    get_status_mock = mocker.patch.object(wifi, '_get_wifi_status_code', return_value=1)
    wifi_mock = mocker.patch.object(wifi, '_disable_wifi')
    intro_msg = 'Disabling WiFi ...'
    echo_mock = mocker.patch('click.echo')

    wifi.turn_off()

    echo_mock.assert_called_once_with(intro_msg)
    get_status_mock.assert_called_once()
    wifi_mock.assert_called_once()


def test_turn_off_wifi_error(wifi, mocker):
    """
        Assert click.echo was called with the correct message passed
        as the argument.
        Assert WiFi._get_wifi_status_code was called and that if
        it returns 0 (disabled), click.echo is called with the correct
        error message.
    """
    get_status_mock = mocker.patch.object(wifi, '_get_wifi_status_code', return_value=0)
    echo_mock = mocker.patch('click.echo')
    expected_err_msg = call('WiFi is already disabled. Skipping.')

    wifi.turn_off()

    get_status_mock.assert_called_once()
    echo_mock.assert_has_calls([expected_err_msg])


def test_print_wifi_status(wifi, mocker):
    """
        Assert Wifi.print_status() correctly retrieves the current
        wireless state, translates it to a colored string and outputs
        the correct message to the user.
    """
    get_status_mock = mocker.patch.object(wifi, '_get_wifi_status_code', return_value=1)
    echo_mock = mocker.patch('click.echo')
    translate_mock = mocker.patch.object(wifi, 'translate_status_code', return_value=colored("enabled", "green"))

    wifi_prop = DBUS_GENERAL_PROPS[wifi.wifi_prop]

    wifi.print_status()

    expected_output = f'WiFi is {colored("enabled", "green")}'
    get_status_mock.assert_called_once()
    translate_mock.assert_called_once_with(wifi_prop, 1)
    echo_mock.assert_called_once_with(expected_output)


def test_enable_wifi(wifi, mocker):
    """
        Assert Wifi._enable_wifi() calls its inherited
        set_object_property method passing it the correct
        proxy, prop name and value (True in this case).
    """
    set_prop_mock = mocker.patch.object(wifi, 'set_object_property')
    wifi._enable_wifi()

    set_prop_mock.assert_called_once_with(
        proxy=wifi.proxy,
        prop_name=wifi.wifi_prop,
        value=True
    )


def test_disable_wifi(wifi, mocker):
    """
        Assert Wifi._disable_wifi() calls its inherited
        set_object_property method passing it the correct
        proxy, prop name and value (False in this case).
    """
    set_prop_mock = mocker.patch.object(wifi, 'set_object_property')

    wifi._disable_wifi()

    set_prop_mock.assert_called_once_with(
        proxy=wifi.proxy,
        prop_name=wifi.wifi_prop,
        value=False
    )


def test_get_wifi_status_code(wifi, mocker):
    get_prop_mock = mocker.patch.object(wifi, 'get_object_property')
    wifi._get_wifi_status_code()
    get_prop_mock.assert_called_once_with(
        proxy=wifi.proxy,
        prop_name=wifi.wifi_prop
    )


# TODO:
def test_format_bitrate(wifi):
    bitrate = 134000.00
    result = wifi._format_bitrate(bitrate)
    assert result == '134 Mbit/s'


def test_get_mode(wifi):
    mode = '4'
    result = wifi._get_mode(mode)
    assert result == 'Mesh'

    mode = '3'
    result = wifi._get_mode(mode)
    assert result == 'Access Point'

    mode = '2'
    result = wifi._get_mode(mode)
    assert result == 'Infra'

    mode = '1'
    result = wifi._get_mode(mode)
    assert result == 'Ad Hoc'

    mode = '0'
    result = wifi._get_mode(mode)
    assert result == 'Unknown'


def test_get_bars(wifi):
    signal = '2'
    result = wifi._get_bars(signal)
    assert result == '*'

    signal = '41'
    result = wifi._get_bars(signal)
    assert result == '**'

    signal = '71'
    result = wifi._get_bars(signal)
    assert result == '***'

    signal = '100'
    result = wifi._get_bars(signal)
    assert result == '****'

    signal = '23420'
    result = wifi._get_bars(signal)
    assert result == '--'


def test_get_ssid(wifi):
    raw_ssid = [0x50, 0x61, 0x72, 0x69, 0x73, 0x20, 0x57, 0x69, 0x2d, 0x46, 0x69]
    result = wifi._get_ssid(raw_ssid)
    assert result == "Paris Wi-Fi"

    raw_ssid = []
    result = wifi._get_ssid(raw_ssid)
    assert result == '--'


def test_extract_access_point_info(wifi, mocker):
    access_point_props = {
        'hello': 'world',
        'temperature': 80,
        'fragile': False,
        'Ssid': [],
        'Mode': '2',
        'MaxBitrate': '982909',
        'Strength': '92',
        'name': 'Karim Benzema',
        'age': 32,
        'current_club': 'Real Madrid CF'
    }
    expected_keys = [
        'ssid', 'mode', 'rate',
        'signal', 'bars', 'dbus_path'
    ]
    get_all_props_mock = mocker.patch.object(
        wifi,
        'get_all_properties',
        return_value=access_point_props
    )

    result = wifi._extract_ap_info('/some/path')

    get_all_props_mock.assert_called_once()
    assert sorted(expected_keys) == sorted(result.keys())


def test_get_all_access_points(wifi, mocker):
    dummy_ap_paths = ['/ap/1', '/ap/2', '/ap/3']
    request_scan = mocker.patch.object(wifi, '_request_scan')
    list_ap_paths = mocker.patch.object(
        wifi,
        '_list_ap_paths',
        return_value=dummy_ap_paths
    )
    extract_ap_info = mocker.patch.object(
        wifi,
        '_extract_ap_info',
        return_value={'name': 'Karim Benzema', 'club': 'Real Madrid CF'}
    )
    dummy_iface = 'dummy_iface'
    dummy_ap_paths_calls = [call(path) for path in dummy_ap_paths]

    result = wifi._get_all_access_points()

    request_scan.assert_called_once()
    list_ap_paths.assert_called_once()
    extract_ap_info.assert_has_calls(dummy_ap_paths_calls)
    assert isinstance(result, list) == True


def test_filter_wifi_devices(wifi):
    device = ('/dev/path', {'Real': 1, 'DeviceType': 2})
    result = wifi._is_device_wifi(device)
    assert result == True

    device = ('/dev/path', {'Real': 0, 'DeviceType': 7623})
    result = wifi._is_device_wifi(device)
    assert result == False


def test_get_wireless_device_path(wifi, mocker):
    dummy_devices = ['/dev/1', '/dev/2']
    get_all_dev_paths = mocker.patch.object(
        wifi,
        '_get_all_devices_paths',
        return_value=dummy_devices
    )
    get_all_props = mocker.patch.object(
        wifi,
        'get_all_properties',
        return_value={}
    )
    is_device_wifi = mocker.patch.object(
        wifi,
        '_is_device_wifi',
        return_value=True
    )

    get_all_props_calls = [call(dev, NM_DEVICE_IFACE) for dev in dummy_devices]
    is_device_wifi_calls = [call((dev, {})) for dev in dummy_devices]

    result = wifi._get_wireless_device_path()

    get_all_dev_paths.assert_called_once()
    get_all_props.assert_has_calls(get_all_props_calls)
    is_device_wifi.assert_has_calls(is_device_wifi_calls)

    assert result == '/dev/1'


def test_connect_to_wifi_existing_conn(wifi, mocker):
    dummy_ap_name = 'MyAwesomeAPName'
    dummy_active_conn_path = '/path/to/active/conn'
    dummy_conn_paths = ['/conn/1', '/conn/2']

    list_all_conns = mocker.patch.object(
        wifi,
        '_list_all_connections',
        return_value=dummy_conn_paths
    )

    get_conn_info_mock = mocker.patch.object(
        wifi,
        '_get_connection_info',
        return_value={'id': dummy_ap_name, 'dbus_path': '/path/to/conn/1'}
    )

    prompt_mock = mocker.patch(
        'click.prompt',
        return_value=dummy_ap_name
    )

    echo_mock = mocker.patch('click.echo')

    activate_existing_connection_mock = mocker.patch.object(
        wifi,
        '_activate_existing_connection',
        return_value=dummy_active_conn_path
    )

    wifi.connect()

    success_msg = echo_mock.call_args[0][0]
    list_all_conns.assert_called_once()
    assert get_conn_info_mock.call_count == len(dummy_conn_paths)
    assert prompt_mock.call_count == 2
    activate_existing_connection_mock.assert_called_once_with('/path/to/conn/1')
    assert dummy_active_conn_path and dummy_ap_name in success_msg


# connect not found
def test_connect_to_wifi_non_existant_conn(wifi, mocker):
    dummy_ap_name_and_pwd = 'MyAwesomeAPNameAndPassword'
    dummy_conn_paths = ['/conn/1', '/conn/2']

    list_all_conns = mocker.patch.object(
        wifi,
        '_list_all_connections',
        return_value=dummy_conn_paths
    )

    get_conn_info_mock = mocker.patch.object(
        wifi,
        '_get_connection_info',
        return_value={'id': 'helloworld', 'dbus_path': '/path/to/conn/1'}
    )

    prompt_mock = mocker.patch(
        'click.prompt',
        return_value=dummy_ap_name_and_pwd
    )

    conn_to_ap_mock = mocker.patch.object(
        wifi,
        '_connect_to_access_point'
    )

    wifi.connect()

    list_all_conns.assert_called_once()
    assert get_conn_info_mock.call_count == len(dummy_conn_paths)
    assert prompt_mock.call_count == 2
    conn_to_ap_mock.assert_called_once_with(
        dummy_ap_name_and_pwd,
        dummy_ap_name_and_pwd
    )


def test_connect_to_ap_unknown(wifi, mocker):
    dummy_ap_name = 'some_ap'
    dummy_ap_pass = 'some_pass'
    dummy_aps = [{'ssid': 'hello'}, {'ssid': 'world'}]
    get_all_aps_mock = mocker.patch.object(
        wifi,
        '_get_all_access_points',
        return_value=dummy_aps
    )
    expected_err = "[Error]: Connection to some_ap impossible. No such access point."
    with pytest.raises(SystemExit) as exc:
        wifi._connect_to_access_point(dummy_ap_name, dummy_ap_pass)

    assert colored(expected_err, "red") == exc.value.args[0]
    get_all_aps_mock.assert_called_once()


def test_connect_to_ap(wifi, mocker):
    wifi_conn_info = {'key': 'value'}
    new_conn_uuid = 'new-conn-uuid-v4'
    dummy_ap_name = 'some_ap'
    dummy_ap_pass = 'some_pass'
    dummy_ap_path = '/path/to/ap'
    dummy_aps = [
        {'ssid': dummy_ap_name, 'dbus_path': dummy_ap_path},
        {'ssid': 'world', 'dbus_path': 'some_path_to_dbus'}
    ]
    get_all_aps_mock = mocker.patch.object(
        wifi,
        '_get_all_access_points',
        return_value=dummy_aps
    )
    generate_wifi_info_mock = mocker.patch.object(
        wifi,
        '_generate_wireless_connection_info',
        return_value=wifi_conn_info
    )
    establish_connection_mock = mocker.patch.object(
        wifi,
        '_establish_connection',
        return_value=new_conn_uuid
    )

    wifi._connect_to_access_point(dummy_ap_name, dummy_ap_pass)

    get_all_aps_mock.assert_called_once()
    generate_wifi_info_mock.assert_called_once_with(dummy_ap_name, dummy_ap_pass)
    establish_connection_mock.assert_called_once_with(wifi_conn_info, dummy_ap_path)


def test_generate_wireless_info(wifi):
    dummy_ap = 'some_ap'
    dummy_pw = 'some_pw'
    info = wifi._generate_wireless_connection_info(
        dummy_ap,
        dummy_pw
    )
    expected_keys = [
        'connection', '802-11-wireless',
        '802-11-wireless-security', 'ipv4', 'ipv6'
    ]
    assert isinstance(info, dbus.Dictionary)
    assert sorted(expected_keys) == sorted(info.keys())

    ipv4 = info['ipv4']
    ipv6 = info['ipv6']

    assert 'method' in ipv4.keys()
    assert 'method' in ipv6.keys()
    assert ipv4['method'] == 'auto'
    assert ipv6['method'] == 'ignore'

    security = info['802-11-wireless-security']
    assert sorted(['key-mgmt', 'auth-alg', 'psk']) == sorted(security.keys())
    assert security['key-mgmt'] == 'wpa-psk'
    assert security['auth-alg'] == 'open'
    assert security['psk'] == dummy_pw

    wifi = info['802-11-wireless']
    assert sorted(['ssid', 'mode']) == sorted(wifi.keys())
    assert wifi['ssid'] == dbus.ByteArray(dummy_ap.encode('utf-8'))
    assert wifi['mode'] == 'infrastructure'

    conn = info['connection']
    assert sorted(['type', 'uuid', 'id']) == sorted(conn.keys())
    assert conn['id'] == dummy_ap
    assert conn['type'] == '802-11-wireless'


def test_exit_loop(wifi, mocker):
    quit_loop_mock = mocker.patch.object(wifi.loop, 'quit')
    echo_mock = mocker.patch('click.echo')

    msg = "Done and dusted !"

    wifi._exit_loop(msg)

    echo_mock.assert_called_once_with(msg)
    quit_loop_mock.assert_called_once()


def test_exit_loop_error(wifi, mocker):
    quit_loop_mock = mocker.patch.object(wifi.loop, 'quit')
    echo_mock = mocker.patch('click.echo')

    msg = "Error !!!!"

    wifi._exit_loop(msg, error=True)

    echo_mock.assert_called_once_with(colored(msg, 'red'))
    quit_loop_mock.assert_called_once()


def test_handle_wifi_state_change(wifi, mocker):
    exit_loop_mock = mocker.patch.object(
        wifi,
        '_exit_loop'
    )
    delete_active_conn = mocker.patch.object(
        wifi,
        '_delete_failed_active_connection'
    )

    new_state = 100
    old_state = 30
    reason = 0
    wifi.ap_name = "Some Ap"
    wifi.ap_uuid = "Some UUID"

    msg = f'Connection to {wifi.ap_name} ({wifi.ap_uuid}) successful !'

    wifi._handle_wifi_state_change(new_state, old_state, reason)

    exit_loop_mock.assert_called_once_with(msg)


def test_handle_wifi_state_change_invalid_pwd(wifi, mocker):
    exit_loop_mock = mocker.patch.object(
        wifi,
        '_exit_loop'
    )
    delete_active_conn = mocker.patch.object(
        wifi,
        '_delete_failed_active_connection'
    )

    new_state = 120
    old_state = 70
    reason = 7

    wifi._handle_wifi_state_change(new_state, old_state, reason)

    msg = '[Error]. Could not connect. Invalid Password.'

    delete_active_conn.assert_called_once()
    exit_loop_mock.assert_called_once_with(msg, error=True)


def test_handle_wifi_state_change_error(wifi, mocker):
    exit_loop_mock = mocker.patch.object(
        wifi,
        '_exit_loop'
    )
    delete_active_conn = mocker.patch.object(
        wifi,
        '_delete_failed_active_connection'
    )

    new_state = 120
    old_state = 70
    reason = 9

    wifi._handle_wifi_state_change(new_state, old_state, reason)

    msg = f'[Error] Could not connect. Reason number: {reason}'

    delete_active_conn.assert_called_once()
    exit_loop_mock.assert_called_once_with(msg, error=True)



from wypy.utils.helpers import format_list, format_table_key
from prettytable import PrettyTable
from termcolor import colored
from wypy.wypy import WyPy
from wypy.utils.constants import (
    NM_BUS_NAME,
    NM_OBJ_PATH,
    NM_IFACE,
    NM_DEVICE_IFACE,
    NM_ACTIVE_CONN_IFACE,
    IP4_CONFIG_IFACE
)
import sys
import click
import dbus
from dbus.exceptions import DBusException


class Device(WyPy):

    def __init__(self):
        super().__init__()
        self.nm = self.bus.get_object(NM_BUS_NAME, NM_OBJ_PATH)
        self.status_table_keys = ['DEVICE', 'TYPE', 'STATE', 'CONNECTION']
        self.status_table = PrettyTable(self.status_table_keys)
        self.status_table.align = 'l'
        self.all_devices = self.get_object_property(
            proxy=self.nm,
            bus_name=NM_IFACE,
            prop_name='AllDevices'
        )
        self.known_device_names = []
        self.details_table = PrettyTable(['PROPERTY', 'VALUE'])
        self.details_table.align = 'l'

    def print_status(self):
        """
        Displays general device status information to the user.
        """
        click.echo('Showing status ...')
        device_status = list(map(self._get_device_status, self.all_devices))  # noqa E501
        sorted_status = sorted(device_status, key=lambda k: (k['connection'], k['type']), reverse=True)  # noqa E501
        rows = list(map(self._create_row, sorted_status))

        for row in rows:
            self.status_table.add_row(row)

        click.echo(self.status_table)

    def list_all(self):
        """
        Prints a table with detailed information for each known device.
        """
        click.echo('Listing all devices ...')
        for device in self.all_devices:
            details = self._get_device_details(device)
            self._fill_details_table(details)

            click.echo(self.details_table)
            self.details_table.clear_rows()

    def print_details(self, ifname):
        """
        Prints a table with details information for the given device.

        Arguments:
            ifname {[string]} -- [the device to display the details for]
        """
        try:
            device_path = self._get_known_device_object_path(ifname)
        except ValueError as exc:
            sys.exit(colored(str(exc), "red"))
        else:
            device_details = self._get_device_details(device_path, show_all=True)  # noqa E501

        self._fill_details_table(device_details)

        click.echo(self.details_table)

    def update_ifname_connection(self, ifname):
        """
        Updates the given device's connection information.
        This is a clone of the "nmcli device update/reapply" command.

        Arguments:
            ifname {string} -- [the device to perform the action on]
        """
        try:
            device_path = self._get_known_device_object_path(ifname)
        except ValueError as exc:
            sys.exit(colored(exc.args, "red"))
        else:
            self._update_connection(device_path, ifname)

    def disconnect(self, ifname):
        """
        Disconnects the given device.

        Arguments:
            ifname {string} -- the device to disconnect
        """
        try:
            device_path = self._get_known_device_object_path(ifname)
        except ValueError as exc:
            sys.exit(colored(exc.args, "red"))
        else:
            self._disconnect_device(device_path, ifname)

    def delete_iface(self, ifname):
        """
        Deletes the given device

        Arguments:
            ifname {string} -- the device to delete
        """
        try:
            device_path = self._get_known_device_object_path(ifname)
        except ValueError as exc:
            sys.exit(colored(exc.args, "red"))
        else:
            self._delete_iface(device_path, ifname)

    def manage(self, ifname, flag=True):
        """
        Sets the given device's Managed property
        to either True of False.

        Arguments:
            ifname {string} -- [the device to manage / unmanage]

        Keyword Arguments:
            flag {bool} -- [manage flag] (default: {True})
        """
        try:
            device_path = self._get_known_device_object_path(ifname)
        except ValueError as exc:
            sys.exit(colored(exc.args, "red"))
        else:
            proxy = self.bus.get_object(NM_BUS_NAME, device_path)
            self.set_object_property(
                proxy,
                bus_name=NM_DEVICE_IFACE,
                prop_name='Managed',
                value=flag
            )

    def autoconnect(self, ifname, flag=True):
        """
        Sets the given device's Autoconnect property
        to either True of False.

        Arguments:
            ifname {string} -- [the device to autoconnect]

        Keyword Arguments:
            flag {bool} -- [enable / disable auto connect] (default: {True})
        """
        try:
            device_path = self._get_known_device_object_path(ifname)
        except ValueError as exc:
            sys.exit(colored(exc.args, "red"))
        else:
            proxy = self.bus.get_object(NM_BUS_NAME, device_path)
            self.set_object_property(
                proxy,
                bus_name=NM_DEVICE_IFACE,
                prop_name='Autoconnect',
                value=flag
            )

    # --------------- #
    # Private methods #
    # --------------- #

    def _get_known_device_object_path(self, ifname):
        """
        Returns the device's d-bus object path.
        If the given device name is not known to the system,
        the program will exit with an error message.

        Arguments:
            device_obj_path {string} -- the device's own dbus object path

        Returns:
            dbus.Interface -- dbus interface for the given device
        """
        known_devices = list(map(self._get_device_status, self.all_devices))  # noqa E501
        self.known_device_names = list(map(lambda x: str(x['name']), known_devices))  # noqa E501

        if ifname not in self.known_device_names:
            err_msg = f"""
            [Error]: Could not perform the desired operation on "{ifname}".
            The requested device does not appear to exist.
            """.replace("  ", "")
            raise ValueError(err_msg)

        device_info = next(filter(lambda x: str(x['name']) == ifname, known_devices), None)  # noqa E501
        device_path = device_info['device_path']

        return device_path

    def _get_device_dbus_interface(self, device_obj_path):
        """
        Creates and returns a dbus interface for the device.

        Arguments:
            device_obj_path {string} -- the device's own dbus object path

        Returns:
            dbus.Interface -- dbus interface for the given device
        """
        device_obj = self.bus.get_object(NM_BUS_NAME, device_obj_path)
        device_iface = dbus.Interface(device_obj, NM_DEVICE_IFACE)
        return device_iface

    def _update_connection(self, device_path, ifname):
        """
        Reapplies the settings on the connection
        associated with the given device.

        Arguments:
            device_path {string} -- the device's own dbus object path
            ifname {string} -- the device's interface name
        """
        device = self._get_device_dbus_interface(device_path)
        try:
            device.Reapply({}, 0, 0)
        except DBusException:
            err_msg = f"""
            [Error]: Could not update active connection info for "{ifname}".
            The device is not activated.
            D-Bus object path: ({device_path})
            """.replace("  ", "")
            sys.exit(colored(err_msg, "red"))
        else:
            msg = f'Successfully updated connection information for "{ifname}"'
            click.echo(msg)

    def _disconnect_device(self, device_path, ifname):
        """
        Gets the device's dbus interface and
        calls the Disconnect method on it.
        If an error occurs, the program exits with an error message.

        Arguments:
            device_path {string} -- the device's own dbus object path
            ifname {string} -- the device's name
        """
        device = self._get_device_dbus_interface(device_path)
        try:
            device.Disconnect()
        except DBusException:
            err_msg = f"""
            [Error]: Could not disconnect {ifname}.
            The device is either not active or is already disconnected.
            D-Bus object path: ({device_path})
            """.replace("  ", "")
            sys.exit(colored(err_msg, "red"))
        else:
            click.echo(f'Device "{ifname}" was successfully disconnected')

    def _delete_iface(self, device_path, ifname):
        """
        Gets the device's dbus interface and
        calls the Delete method on it.
        If an error occurs, the program exits with an error message.

        Arguments:
            device_path {string} -- the device's own dbus object path
            ifname {string} -- the device's name
        """
        device = self._get_device_dbus_interface(device_path)
        try:
            device.Delete()
        except DBusException as exc:
            err_msg = exc.get_dbus_message()
            err_msg = f'[Error]: Device - "{ifname}"\n{err_msg}'
            sys.exit(colored(err_msg, "red"))
        else:
            click.echo(f"Successfully deleted {ifname}")

    def _get_device_status(self, obj_path):
        """
        Retrieve general status related information
        about a device.

        Arguments:
            obj_path {string} -- the device's own d-bus object path

        Returns:
            dict -- the device's status information
        """
        dev_props = self.get_all_properties(obj_path, NM_DEVICE_IFACE)
        self._stringify_dbus_values(dev_props)

        dev_name = dev_props.get('Interface', 'Unknown')
        dev_type = dev_props.get('DeviceType', 'Unknown')
        dev_state = dev_props.get('State', 'Unknown')
        dev_conn = dev_props.get('ActiveConnection', '--')

        return {
            'name': dev_name,
            'type':  self.translate_device_type(dev_type),
            'device_status': self.translate_device_state(dev_state),
            'connection': self._get_connection_name(dev_conn),
            'state':  dev_state,
            'connection_path': dev_conn,
            'device_path': obj_path
        }

    def _create_row(self, device_details):
        """
        Maps over the values of `device_details` and
        applies the correct color to each value.

        Arguments:
            device_details {dict} -- the device's
            *detailed* information (ipv4, dns ...)

        Returns:
            list -- colored values
        """
        state = int(device_details['state'])
        values = device_details.values()
        row_color = self.get_device_state_row_color(state)

        del device_details['device_path']
        del device_details['connection_path']
        del device_details['state']

        return list(map(lambda val: colored(val, row_color), values))

    def _get_connection_name(self, connection_path):
        """
        Retrives the name of the given connection.
        If an error occurs, the program exits with an erorr message.

        Arguments:
            connection_path {string} -- the connection's own d-bus object path

        Returns:
            string -- the connection's name
        """
        try:
            props = self.get_all_properties(
                connection_path,
                NM_ACTIVE_CONN_IFACE
            )
        except DBusException:
            return '--'
        except Exception as e:
            click.echo('An error occured', str(e))
        return props['Id']

    def _stringify_dbus_values(self, _dict):
        # loop through the values and cast them appropriately
        # move that method to WyPy
        map(lambda val: str(val), _dict.values())

    def _get_device_details(self, device_obj, show_all=False):
        """
        Retrives detailed information about the given device.
        By default, the program only returns the status information
        along with an additional 'Mtu' property.
        If the show_all flag is enabled, WyPy will gather even more
        details.

        Arguments:
            device_obj {string} -- the device's own d-bus object path

        Keyword Arguments:
            show_all {bool} -- whether to
            show all details or not (default: {False})

        Returns:
            dict -- the device's details
        """
        dev_props = self.get_all_properties(device_obj, NM_DEVICE_IFACE)
        self._stringify_dbus_values(dev_props)
        dev_status = self._get_device_status(device_obj)

        general_dev_info = {
            'mtu': dev_props.get('Mtu', 'Unknown'),
        }
        result = dict(dev_status, **general_dev_info)

        if show_all:
            ip4_path = dev_props['Ip4Config']
            ip4props = self.get_all_properties(ip4_path, IP4_CONFIG_IFACE)

            addresses = format_list(ip4props['AddressData'], key='address')
            dns = format_list(ip4props['NameserverData'], key='address')
            gateway = ip4props['Gateway'] if ip4props['Gateway'] != '' else '--'  # noqa E501
            domains = ' / '.join(ip4props['Domains']) if len(ip4props['Domains']) != 0 else '--'  # noqa E501

            ip_info = {
                'ipv4_addresses': addresses,
                'ipv4_dns': dns,
                'ipv4_gateway': gateway,
                'ipv4_domains': domains
            }

            result = dict(result, **ip_info)

        return result

    def _fill_details_table(self, data):
        """
        Iterates over `data`'s items
        Uppercases / colors the key in yellow
        Passes the newly formatted key and its associated value to
        PrettyTable.add_row()

        Arguments:
            data {dict} -- the device's properties
        """
        for k, v in data.items():
            k = format_table_key(k)
            self.details_table.add_row([colored(k, "yellow"), v])



from gi.repository import GObject  # noqa F401
from termcolor import colored
from prettytable import PrettyTable
from dbus.exceptions import DBusException
from wypy.utils.constants import (
    NM_CONNECTION_IFACE,
    NM_BUS_NAME,
    NM_OBJ_PATH,
    NM_SETTINGS_IFACE,
    NM_SETTINGS_OBJ_PATH,
    DBUS_GENERAL_PROPS,
    NM_IFACE,
    NM_DEVICE_IFACE,
    NM_WIRELESS_IFACE,
    NM_ACCESS_POINT_IFACE,
)
from wypy.wypy import WyPy
import dbus, click, sys, time, uuid  # noqa E401


class WiFi(WyPy):

    def __init__(self):
        super().__init__()
        self.proxy = self.bus.get_object(NM_BUS_NAME, NM_OBJ_PATH)
        self.wifi_prop = 'WirelessEnabled'
        self.status_table = PrettyTable(['SSID', 'MODE', 'RATE', 'SIGNAL', 'BARS'])  # noqa E501
        self.status_table.align = 'l'
        self.status_table.border = False
        self.status_table.sortby = 'SIGNAL'
        self.status_table.left_padding_width = 0
        self.status_table.right_padding_width = 8
        self.settings_obj = self.bus.get_object(
            NM_BUS_NAME,
            NM_SETTINGS_OBJ_PATH
        )
        self.settings_iface = dbus.Interface(
            self.settings_obj,
            NM_SETTINGS_IFACE
        )
        try:
            self.wifi_dev_path = self._get_wireless_device_path()
        except Exception:
            pass
        else:
            self.wifi_dev_obj = self.bus.get_object(
                NM_BUS_NAME,
                self.wifi_dev_path
            )
            self.wifi_iface = dbus.Interface(
                self.wifi_dev_obj,
                NM_WIRELESS_IFACE
            )
        finally:
            self.loop = GObject.MainLoop()

    def list_access_points(self):
        """
        Lists all visible access points.
        """
        click.echo("Scanning for available access points ...")
        access_points = self._get_all_access_points()
        rows = list(map(self._create_row, access_points))

        for row in rows:
            self.status_table.add_row(row)

        click.echo(self.status_table)

    def rescan(self):
        """
        Rescans the network for ( potentially ) newly added access points.
        """
        click.echo('Performing rescan ...')
        self._request_scan()
        click.echo('Done !')

    def connect(self):
        """
            # - get all connections
            - is there a matching known connection with that name ?
                - if so is it active ?
                    - if it's not active activate it
            - if there is no matching connection
            - add and activate one passing it the type,
            name, and password (as per the example on github)
        """
        conns_paths = self._list_all_connections()
        conns_info = list(map(self._get_connection_info, conns_paths))
        conn_names = list(map(lambda conn: str(conn['id']), conns_info))

        access_point_name = click.prompt('SSID name')
        access_point_password = click.prompt('Password')

        if access_point_name in conn_names:
            _filter = lambda conn: conn['id'] == access_point_name  # noqa E731
            conn_to_activate = next(filter(_filter, conns_info))
            conn_path = conn_to_activate['dbus_path']
            active_conn_path = self._activate_existing_connection(conn_path)
            click.echo(f'Connection to {access_point_name} successful ! (D-Bus path: {active_conn_path})')  # noqa E501
        else:
            self._connect_to_access_point(
                access_point_name,
                access_point_password
            )

    def turn_on(self):
        """
        Enables the network card's wireless capability.
        """
        click.echo('Enabling WiFi ...')
        wifi_enabled = self._get_wifi_status_code()
        if not wifi_enabled:
            self._enable_wifi()
        else:
            click.echo('WiFi is already enabled. Skipping.')

    def turn_off(self):
        """
        Disables the network card's wireless capability.
        """
        click.echo('Disabling WiFi ...')
        wifi_enabled = self._get_wifi_status_code()
        if wifi_enabled:
            self._disable_wifi()
        else:
            click.echo('WiFi is already disabled. Skipping.')

    def print_status(self):
        """
        Prints the current wifi status to the user.
        """
        status_code = self._get_wifi_status_code()
        prop = DBUS_GENERAL_PROPS[self.wifi_prop]
        status = self.translate_status_code(prop, status_code)
        click.echo(f'WiFi is {status}')

    #   ---------------
    #
    #   Private Methods
    #
    #   ---------------

    def _connect_to_access_point(self, ap_name, ap_pwd):
        """
        Establishes a wireless connection to the
        provided access point.

        Arguments:
            ap_name {string} -- the access point's name (SSID)
            ap_pwd {string} -- the access point's password
        """

        all_access_points = self._get_all_access_points()

        _filter = lambda ap: ap['ssid'] == ap_name  # noqa E731
        ap_to_connect_to = next(filter(_filter, all_access_points), None)

        if ap_to_connect_to is not None:
            ap_path = ap_to_connect_to['dbus_path']
            conn_info = self._generate_wireless_connection_info(
                ap_name,
                ap_pwd
            )
            click.echo('Establishing connection. Please hold tight ...')
            self._establish_connection(conn_info, ap_path)
        else:
            msg = f'[Error]: Connection to {ap_name} impossible. No such access point.'  # noqa E501
            sys.exit(colored(msg, "red"))

    def _handle_wifi_state_change(self, new_state, old_state, reason):
        """
        This is the StateChanged signal handler for the Device
        D-Bus interface.
        It processes the new state, and the potential
        reason in case of failure.

        Arguments:
            new_state {int} -- the device's new state
            old_state {int} -- the device's old state
            reason {int} -- the reason why the device failed to connect.
        """
        if new_state == 100:
            msg = f'Connection to {self.ap_name} ({self.ap_uuid}) successful !'
            self._exit_loop(msg)

        if new_state == 120:
            if reason == 7:
                msg = '[Error]. Could not connect. Invalid Password.'
            else:
                msg = f'[Error] Could not connect. Reason number: {reason}'
            self._delete_failed_active_connection()
            self._exit_loop(msg, error=True)

    def _delete_failed_active_connection(self):
        """
        Deletes the active connection if the attempt at connecting
        failed.
        This is done to prevent WyPy to accidentally connect the user to
        an inactive access point.
        """
        conn_path = self.settings_iface.GetConnectionByUuid(
            self.ap_uuid
        )
        conn_obj = self.bus.get_object(NM_BUS_NAME, conn_path)
        conn_iface = dbus.Interface(conn_obj, NM_CONNECTION_IFACE)
        conn_iface.Delete()

    def _establish_connection(self, conn, ap_path):
        """
        Adds and activate the connection using the provided info and
        the access point's path.

        Subscribes to the StateChanged signal on the wireless
        device's proxy.

        Runs the GObject main loop until the new state is
        processed by the handler method.

        Arguments:
            conn {dict} -- the connection's information
            ap_path {string} -- the access point's own d-bus object path
        """

        nm = dbus.Interface(self.proxy, NM_IFACE)
        settings, active_conn = nm.AddAndActivateConnection(
            conn,
            self.wifi_dev_path,
            ap_path
        )
        try:
            self.wifi_dev_obj.connect_to_signal(
                'StateChanged',
                self._handle_wifi_state_change,
                NM_DEVICE_IFACE,
            )
            self.ap_name = conn['connection']['id']
            self.ap_uuid = conn['connection']['uuid']
        except DBusException as exc:
            msg = exc.get_dbus_message()
            sys.exit(msg)

        self.loop.run()

    def _generate_wireless_connection_info(self, ap_name, ap_pwd):
        """
        Produces a dictionary containing all the necessary information
        NetworkManager requires to establish a WPA wireless connection.

        Arguments:
            ap_name {string} -- the name of the access point to connect to
            ap_pwd {string} -- the password for the access point

        Returns:
            dict -- the connection info
        """
        conn = dbus.Dictionary({
            'type': '802-11-wireless',
            'uuid': str(uuid.uuid4()),
            'id': ap_name
        })

        settings_wifi = dbus.Dictionary({
            'ssid': dbus.ByteArray(ap_name.encode('utf-8')),
            'mode': 'infrastructure'
        })

        settings_wifi_security = dbus.Dictionary({
            'key-mgmt': 'wpa-psk',
            'auth-alg': 'open',
            'psk': ap_pwd
        })

        settings_ipv4 = dbus.Dictionary({'method': 'auto'})
        settings_ipv6 = dbus.Dictionary({'method': 'ignore'})

        return dbus.Dictionary({
            'connection': conn,
            '802-11-wireless': settings_wifi,
            '802-11-wireless-security': settings_wifi_security,
            'ipv4': settings_ipv4,
            'ipv6': settings_ipv6
        })

    def _activate_existing_connection(self, conn_path):
        """
        Activates an existing connection by calling the
        ActivateConnection method available on the main NetworkManager
        interface.
        If the connection attempt fails, the program exits with
        the error message returned by D-Bus.

        Arguments:
            conn_path {string} -- the connection's own d-bus path

        Returns:
            string -- the newly activated connection's object path
        """
        self.nm = dbus.Interface(self.proxy, NM_IFACE)
        try:
            active_conn = self.nm.ActivateConnection(
                conn_path,
                self.wifi_dev_path,
                '/'
            )
        except DBusException as exc:
            msg = exc.get_dbus_message()
            sys.exit(msg)
        else:
            return active_conn

    def _get_connection_info(self, conn_path):
        conn_obj = self.bus.get_object(NM_BUS_NAME, conn_path)
        conn_iface = dbus.Interface(conn_obj,  NM_CONNECTION_IFACE)
        conn_info = conn_iface.GetSettings()['connection']
        conn_info['dbus_path'] = conn_path
        return conn_info

    def _list_all_connections(self):
        return self.settings_iface.ListConnections()

    def _enable_wifi(self):
        """
        Sets the 'WirelessEnabled' property to False
        on the main NetworkManager d-bus interface.
        """
        self.set_object_property(
            proxy=self.proxy,
            prop_name=self.wifi_prop,
            value=True
        )

    def _disable_wifi(self):
        """
        Sets the 'WirelessEnabled' property to False
        on the main NetworkManager d-bus interface.
        """
        self.set_object_property(
            proxy=self.proxy,
            prop_name=self.wifi_prop,
            value=False
        )

    def _get_wifi_status_code(self):
        """
        Retrieves the current status code for
        wireless connectivity.

        Returns:
            string -- wifi status code
        """
        return self.get_object_property(
            proxy=self.proxy,
            prop_name=self.wifi_prop
        )

    def _create_row(self, ap_data):
        """
        Creates a PrettyTable row using the access point's
        values.
        This function will color the row according to the
        value of the 'signal' key.

        Arguments:
            ap_data {dict} -- the access point's infomation

        Returns:
            list -- [description]
        """
        del ap_data['dbus_path']

        signal = ap_data['signal']
        values = ap_data.values()

        if signal in range(0, 31):
            color = "cyan"
        elif signal in range(30, 61):
            color = "magenta"
        elif signal in range(60, 81):
            color = "yellow"
        elif signal in range(80, 101):
            color = "green"
        else:
            color = "white"

        return list(map(lambda val: colored(val, color), values))

    def _get_wireless_device_path(self):
        """
        Gets the d-bus object path for the wireless device(s)
        _Note_: in this current implementation, we return the object path
        for the first wireless device returned. Support for multiple devices
        will be added in the coming versions.

        Returns:
            string -- the d-bus object path for the wireless device
        """
        devices_paths = self._get_all_devices_paths()
        all_devices = list(map(lambda dev: self.get_all_properties(dev, NM_DEVICE_IFACE), devices_paths))  # noqa E501
        wireless_devices = list(filter(self._is_device_wifi, zip(devices_paths, all_devices)))  # noqa E501

        try:
            wifi_dev_path = wireless_devices[0]  # extract the first real wifi device that's returned  # noqa E501
        except Exception:
            err_msg = '[Error]: No wireless device found'
            sys.exit(colored(err_msg, "red"))
        else:
            return wifi_dev_path[0]

    def _get_all_devices_paths(self):
        """
        Gets object path for all available devices.

        Returns:
            list -- list of object paths for all devices
        """
        nm = dbus.Interface(self.proxy, NM_IFACE)
        return nm.GetAllDevices()

    def _is_device_wifi(self, device):
        """
        Checks if the provided device is both real and wireless enabled.
        This is done by checking the 'Real' and 'DeviceType' properties on
        the given device.

        Arguments:
            device {device} -- the device tuple containing the dbus object
            path at index [0] and the actual dictionary of info at index [1]

        Returns:
            bool -- whether the device is wireless
        """
        device_info = device[1]
        real, device_type = device_info['Real'], device_info['DeviceType']
        return bool(real) and device_type == 2

    def _get_all_access_points(self):
        """
        Gets information for all visible access points
        on the network.

        Arguments:
            wifi_iface {dbus.Interface} -- the wireless interface
            to call GetAllAccessPoints on

        Returns:
            list -- access points dictionaries
        """

        try:
            self._request_scan()
        except SystemExit:
            time.sleep(4)

        access_points_paths = self._list_ap_paths()
        access_points = list(map(self._extract_ap_info, access_points_paths))
        return access_points

    def _list_ap_paths(self):
        """
        Calls GetAllAccessPoints on the wireless d-bus
        interface provided by NetworkManager.

        Arguments:
            wifi_iface {dbus.Interface} -- the wireless interface
            to call GetAllAccessPoints on

        Returns:
            list -- list of access points object paths
        """
        return self.wifi_iface.GetAllAccessPoints()

    def _request_scan(self):
        """
        Calls RequestScan on the wireless d-bus interface
        provided by NetworkManager.

        Arguments:
            wifi_iface {dbus.Interface} -- the wireless interface
            to call RequestScan on
        """
        try:
            self.wifi_iface.RequestScan({})
        except DBusException as exc:
            msg = exc.get_dbus_message()
            err = "Scanning not allowed immediately following previous scan"
            if msg == err:
                sys.exit(colored(f"[Error]: {msg}", "red"))

    def _extract_ap_info(self, ap_path):
        """
        Extract necessary access point properties.
        Namely:
            - the ssid
            - the mode
            - the bitrate
            - the signal strength
            - the bars (visual representation of signal strength)
            - the dbus object path

        Arguments:
            ap_path {string} -- the access point's own d-bus object path

        Returns:
            dict -- access point info
        """
        props = self.get_all_properties(ap_path, NM_ACCESS_POINT_IFACE)
        return {
            'ssid': self._get_ssid(props.get('Ssid')),
            'mode': self._get_mode(props.get('Mode', '--')),
            'rate': self._format_bitrate(props.get('MaxBitrate')),
            'signal': int(props.get('Strength', '--')),
            'bars': self._get_bars(props.get('Strength')),
            'dbus_path': ap_path

        }

    def _get_ssid(self, ssid_byte_list):
        """
        Translates ssid from byte list to string.
        Arguments:
            ssid_byte_list {list} -- the raw ssid in byte form

        Returns:
            string -- the ssid name
        """
        if len(ssid_byte_list) == 0:
            return "--"

        ssid = ""
        for char in ssid_byte_list:
            ssid = ssid + chr(char)

        return ssid

    def _get_bars(self, signal):
        """
        Visually describes signal strength by assigning
        it to a number or bars (*)

        Arguments:
            signal {string} -- the device's signal strength

        Returns:
            string -- the number or bars
        """
        signal = int(signal)
        if signal in range(0, 30):
            return "*"
        elif signal in range(30, 60):
            return "**"
        elif signal in range(60, 80):
            return "***"
        elif signal in range(80, 101):
            return "****"
        else:
            return "--"

    def _get_mode(self, mode):
        """
        Convert device mode integer to its string equivalent.

        Arguments:
            mode {str} -- the device mode

        Returns:
            string -- the corresponding human readable mode
        """
        mode = int(mode)
        if mode == 4:
            return "Mesh"

        if mode == 3:
            return "Access Point"

        if mode == 2:
            return "Infra"

        if mode == 1:
            return "Ad Hoc"

        if mode == 0:
            return "Unknown"

    def _format_bitrate(self, bitrate):
        """
        Convert bitrate value into a human readable string

        Arguments:
            bitrate {float} -- bitrate

        Returns:
            string -- human readable bitrate
        """
        bitrate = int(bitrate) // 1000
        return f"{bitrate} Mbit/s"

    def _exit_loop(self, msg, error=False):
        """
        Exit the main D-bus loop.
        If the error flag is set to true, the msg is colored red.

        Arguments:
            msg {string} -- the message to display the user

        Keyword Arguments:
            error {bool} -- if the message is an error (default: {False})
        """
        if error:
            msg = colored(msg, 'red')
        click.echo(msg)
        self.loop.quit()



import dbus
from wypy.utils.constants import NM_BUS_NAME
from termcolor import colored


class WyPy(object):

    def __init__(self):
        self.bus = dbus.SystemBus()
        self.bus_name = NM_BUS_NAME

    def set_object_property(
        self,
        proxy=None,
        bus_name=None,
        prop_name=None,
        value=None,
        iface=dbus.PROPERTIES_IFACE
    ):
        _bus_name = bus_name if bus_name else self.bus_name
        return proxy.Set(
            _bus_name,
            prop_name,
            value,
            dbus_interface=iface
        )

    def get_object_property(
        self,
        proxy=None,
        bus_name=None,
        prop_name=None,
        iface=dbus.PROPERTIES_IFACE
    ):
        _bus_name = bus_name if bus_name else self.bus_name
        return proxy.Get(
            _bus_name,
            prop_name,
            dbus_interface=iface
        )

    def get_all_properties(self, object_path, iface_name):
        proxy = self.bus.get_object(self.bus_name, object_path)
        iface = dbus.Interface(proxy, dbus.PROPERTIES_IFACE)
        return iface.GetAll(iface_name)

    def translate_status_code(self, prop, code):

        if prop == 'CONNECTIVITY':

            if code == 0:
                return colored('unknown', 'red')

            if code == 1:
                return colored('none', 'red')

            if code == 2:
                return colored('portal', 'yellow')

            if code == 3:
                return colored('limited', 'green')

            if code == 4:
                return colored('full', 'green')

        if prop == 'STATE':
            if code == 0:
                return colored('unknown', 'red')

            if code == 10:
                return colored('asleep', 'yellow')

            if code == 20:
                return colored('disconnected', 'red')

            if code == 30:
                return colored('disconnecting', 'red')

            if code == 40:
                return colored('connecting', 'yellow')

            if code == 50:
                return colored('connected (local)', 'green')

            if code == 60:
                return colored('connected (site)', 'green')

            if code == 70:
                return colored('connected', 'green')

        if prop in ['NETWORKING', 'WIFI', 'WIFI-HW', 'WWAN', 'WWAN-HW']:
            if code == 1:
                return colored('enabled', 'green')
            if code == 0:
                return colored('disabled', 'red')

    def translate_device_type(self, _type):
        if _type == 30:
            return 'WiFi P2P'

        if _type == 29:
            return 'WireGuard'

        if _type == 28:
            return '6LoWPAN'

        if _type == 27:
            return 'WPAN'

        if _type == 26:
            return 'Open vSwitch Bridge'

        if _type == 25:
            return 'Openv Switch Port'

        if _type == 24:
            return 'Openv Switch Interface'

        if _type == 23:
            return 'PPP'

        if _type == 22:
            return 'Dummy'

        if _type == 21:
            return 'MACsec'

        if _type == 20:
            return 'VETH'

        if _type == 19:
            return 'VXLAN'

        if _type == 18:
            return 'MACVLAN'

        if _type == 17:
            return 'IP tunnel'

        if _type == 16:
            return 'TUN/TAP'

        if _type == 15:
            return 'Team Master'

        if _type == 14:
            return 'Generic / Loopback'

        if _type == 13:
            return 'Bridge'

        if _type == 12:
            return 'ADSL Modem'

        if _type == 11:
            return 'VLAN'

        if _type == 10:
            return 'Bond'

        if _type == 9:
            return 'IP Infiniband'

        if _type == 8:
            return 'Modem'

        if _type == 7:
            return 'WiMax'

        if _type == 6:
            return 'OLPC XO Mesh'

        if _type == 5:
            return 'Bluetooth'

        if _type in [4, 3]:
            return 'Unused'

        if _type == 2:
            return 'WiFi'

        if _type == 1:
            return 'Ethernet'

        if _type == 0:
            return 'Unknown'

    def translate_device_state(self, status):
        if status == 120:
            return 'failed'
        if status == 110:
            return 'deactivating'
        if status == 100:
            return "connected"
        if status == 90:
            return "secondaries"
        if status == 80:
            return "ip check"
        if status == 70:
            return "ip config"
        if status == 60:
            return "auth required"
        if status == 50:
            return "configuring"
        if status == 40:
            return "preparing"
        if status == 30:
            return "disconnected"
        if status == 20:
            return "unavailable"
        if status == 10:
            return "unmanaged"

    def get_device_state_row_color(self, status):
        if status == 100:
            return "green"
        elif status in [120, 110, 30]:
            return "red"
        else:
            return "yellow"



import click
from wypy.wypy import WyPy
from wypy.utils.constants import (
    NM_SETTINGS_IFACE,
    NM_SETTINGS_OBJ_PATH,
    NM_OBJ_PATH,
    NM_IFACE,
    DBUS_GENERAL_PROPS
)


class General(WyPy):

    def __init__(self):
        super().__init__()
        self.status_properties = [
            'State', 'Connectivity', 'WirelessEnabled',
            'WirelessHardwareEnabled', 'WwanEnabled',
            'WwanHardwareEnabled'
        ]

    def show_status(self):
        """
        Display the general status of NetworkManager
        """
        status_info = self._get_status_info()
        click.echo('WyPy - General status report \n')
        for prop in status_info:
            name = prop['name']
            status_code = prop['status_code']
            click.echo(f'{name}: {self.translate_status_code(name, status_code)} \n')  # noqa: E501

    def get_hostname(self):
        """
        Get the 'Hostname' property on the main NetworkManager
        d-bus interface and echo it to the user.
        """
        proxy = self.bus.get_object(self.bus_name, NM_SETTINGS_OBJ_PATH)
        hostname = self.get_object_property(
            proxy=proxy,
            bus_name=NM_SETTINGS_IFACE,
            prop_name='Hostname'
        )
        click.echo(f'Hostname: {hostname}')

    #   ---------------
    #
    #   Private Methods
    #
    #   ---------------

    def _get_status_info(self):
        """
        Get all properties available on the main NetworkManager
        dbus interface. Filter only those listed in self.status_properties.

        Returns:
            status_info -- list of dictionaries each containing
            information on the properties listed in self.status_properties.
        """
        props = self.get_all_properties(
            object_path=NM_OBJ_PATH,
            iface_name=NM_IFACE
        )

        status_info = [
            {
                'name': DBUS_GENERAL_PROPS[str(prop)],
                'status_code': int(props[prop])
            }
            for prop in self.status_properties
        ]

        return status_info



from termcolor import colored
from dbus.proxies import ProxyObject


def test_get_connectivity_state(network, mocker):
    """
    Assert that Network.get_connectivity_state does the following:
        - calls self.bus.get_object_property
        - calls click.echo with the correct string
    """
    expected_msg = f'Connectivity state: { colored("full", "green") }'

    get_prop_mock = mocker.patch.object(network, 'get_object_property', return_value=4)
    echo_mock = mocker.patch('click.echo')

    network.get_connectivity_state()

    mock_prop_name = get_prop_mock.call_args[1]['prop_name']
    mock_proxy_obj = get_prop_mock.call_args[1]['proxy']

    echo_mock.assert_called_once()
    echo_mock.assert_called_once_with(expected_msg)

    assert mock_prop_name == 'Connectivity'
    assert isinstance(mock_proxy_obj, ProxyObject)


def test_networking_turn_on(network, mocker):
    """
        - assert Network.get_networking_status was called
        - assert Network._enable_networking was called
    """

    get_status_mock = mocker.patch.object(network, '_get_networking_status', return_value=0)
    enable_network_mock = mocker.patch.object(network, '_enable_networking')

    network.turn_on()
    get_status_mock.assert_called_once()
    enable_network_mock.assert_called_once()


def test_networking_turn_on_error(network, mocker):
    """
        This test case ensures that an error message is printed
        to the user when they attempt to enable networking and
        it's already on.

        - assert Network.get_networking_status was called
        - assert click.echo was called with the expected error message
    """

    echo_mock = mocker.patch('click.echo')
    get_status_mock = mocker.patch.object(network, '_get_networking_status', return_value=1)

    expected_msg = 'Networking is already enabled. Skipping.'
    network.turn_on()
    get_status_mock.assert_called_once()
    echo_mock.assert_called_once_with(expected_msg)

def test_networking_turn_off(network, mocker):
    """
        - assert Network.get_networking_status was called
        - assert Network._disable_networking was called
    """

    get_status_mock = mocker.patch.object(network, '_get_networking_status', return_value=1)
    disable_network_mock = mocker.patch.object(network, '_disable_networking')

    network.turn_off()
    get_status_mock.assert_called_once()
    disable_network_mock.assert_called_once()


def test_networking_turn_off_error(network, mocker):
    """
        - assert Network.get_networking_status was called (return 0)
        - assert click.echo was called with the message
    """
    echo_mock = mocker.patch('click.echo')
    get_status_mock = mocker.patch.object(network, '_get_networking_status', return_value=0)
    expected_msg = 'Networking is already disabled. Skipping.'

    network.turn_off()

    get_status_mock.assert_called_once()
    echo_mock.assert_called_once_with(expected_msg)


def test_get_networking_status(network, mocker):
    """
        - assert WyPy.get_object_property was called with
        'NetworkingEnabled' passed as an argument along
        with self.proxy
    """

    get_prop_mock = mocker.patch.object(network, 'get_object_property')

    network._get_networking_status()
    get_prop_mock.assert_called_once_with(
        prop_name='NetworkingEnabled',
        proxy=network.proxy
    )



from click.testing import CliRunner
from wypy.cli import cli
from wypy.utils.constants import VERSION


def test_wypy_version():
    runner = CliRunner()
    result = runner.invoke(cli, ['--version'])
    assert result.exit_code == 0
    assert result.output == f'WyPy - Version {VERSION}\n'



from wypy.utils.helpers import (
    is_valid_uuid,
    format_list,
    format_table_key
)


def test_format_list():
    addresses = [
        {'address': '192.168.0.1'},
        {'address': '192.168.0.2'},
        {'address': '192.168.0.3'},
    ]
    result = format_list(addresses)
    expected_result = '192.168.0.1 - 192.168.0.2 - 192.168.0.3'
    assert result == expected_result

    domains = [
        {'age': 23, 'name': 'karim'},
        {'age': 24, 'name': 'hatem'},
        {'age': 25, 'name': 'samir'}
    ]
    result = format_list(domains, key='name')
    expected_result = 'karim - hatem - samir'
    assert result == expected_result

    empty_list = []
    result = format_list(empty_list)
    assert result == '--'


def test_format_table_key():
    key = 'ipv4_address'
    assert format_table_key(key) == 'IPV4 ADDRESS'

    key = 'hello'
    assert format_table_key(key) == 'HELLO'


def test_is_valid_uuid():
    _str = 'kwjkewe'
    assert is_valid_uuid(_str) == False

    _str = '67a548de-8383-4330-abf8-ce60544c5366'
    assert is_valid_uuid(_str) == True



import click
import dbus
from wypy.wypy import WyPy
from wypy.utils.constants import (
    DBUS_GENERAL_PROPS,
    NM_IFACE,
    NM_OBJ_PATH,
    NM_BUS_NAME
)


class Network(WyPy):

    def __init__(self):
        super().__init__()
        self.proxy = self.bus.get_object(NM_BUS_NAME, NM_OBJ_PATH)
        self.iface = dbus.Interface(self.proxy, NM_IFACE)
        self.connectivity_prop_name = 'Connectivity'

    def get_connectivity_state(self):
        """
        Retrieve general connectivity information from dbus.
        """
        status_code = self._get_connectivity_status_code()
        prop = DBUS_GENERAL_PROPS[self.connectivity_prop_name]
        connectivity_state = self.translate_status_code(prop, status_code)
        self._print_connectivity_state(connectivity_state)

    def turn_on(self):
        """
        Enable networking on NetworkManager.
        A check is performed before hand.
        WyPy will exit with an error message if networking is already enabled.
        """
        networking_enabled = self._get_networking_status()
        if not networking_enabled:
            self._enable_networking()
        else:
            click.echo('Networking is already enabled. Skipping.')

    def turn_off(self):
        """
        Disable networking on NetworkManager.
        A check is performed before hand.
        WyPy will exit with an error message if networking is already disabled.
        """
        networking_enabled = self._get_networking_status()
        if networking_enabled:
            self._disable_networking()
        else:
            click.echo('Networking is already disabled. Skipping.')

    def check_connectivity(self):
        click.echo('Performing connectivity check ...')
        """
        Forces NetworkManager to perform a connectivity check
        """
        status_code = self._check_connectivity_state()
        prop = DBUS_GENERAL_PROPS[self.connectivity_prop_name]
        connectivity_state = self.translate_status_code(prop, status_code)
        self._print_connectivity_state(connectivity_state)

    #   ---------------
    #
    #   Private Methods
    #
    #   ---------------

    def _get_connectivity_status_code(self):
        """
        Retrive the 'Connectivity' property available
        on NetworkManager's main d-bus interface.

        Returns:
            int -- the current network connectivity status code
        """
        status_code = self.get_object_property(
            proxy=self.proxy,
            prop_name=self.connectivity_prop_name
        )
        return int(status_code)

    def _print_connectivity_state(self, status):
        """
        Echo the current connectivity state to the user.

        Arguments:
            status {string} -- the current connectivity status
        """
        click.echo(f'Connectivity state: {status}')

    def _check_connectivity_state(self):
        """
        Calls the CheckConnectivity method on the NetworkManager's
        main d-bus Interface.
        Returns:
            int -- the current networking connectivity state
        """
        state = self.iface.CheckConnectivity()
        return state

    def _enable_networking(self):
        """
        Calls the `Enable` method available on the main
        NetworkManager d-bus interface.

        interface: org.freedesktop.NetworkManager
        """
        self.iface.Enable(True)

    def _disable_networking(self):
        """
        Calls the `Disable` method available on the main
        NetworkManager d-bus interface.

        interface: org.freedesktop.NetworkManager
        """
        self.iface.Enable(False)

    def _get_networking_status(self):
        """
        Retrieve the `NetworingEnabled` property on the main
        NetworkManager d-bus Interface.

        Returns:
            [int] -- [the networking status code]
        """
        status_code = self.get_object_property(
            proxy=self.proxy,
            prop_name='NetworkingEnabled'
        )
        return int(status_code)



from termcolor import colored


def test_translate_status_code_connectivity(wypy):
    """
    Assert General.translate_status_code method returns the
    correct string for the CONNECTIVITY property.
    """
    prop = 'CONNECTIVITY'

    result = wypy.translate_status_code(prop, 4)
    assert result == colored('full', 'green')

    result = wypy.translate_status_code(prop, 3)
    assert result == colored('limited', 'green')

    result = wypy.translate_status_code(prop, 2)
    assert result == colored('portal', 'yellow')

    result = wypy.translate_status_code(prop, 1)
    assert result == colored('none', 'red')

    result = wypy.translate_status_code(prop, 0)
    assert result == colored('unknown', 'red')


def test_translate_status_code_wifi(wypy):
    """
    Assert General.translate_status_code returns the
    correct string for the WIFI property.
    """
    prop = 'WIFI'

    result = wypy.translate_status_code(prop, 1)
    assert result == colored('enabled', 'green')

    result = wypy.translate_status_code(prop, 0)
    assert result == colored('disabled', 'red')


def test_translate_status_code_state(wypy):
    """
    Assert General.translate_status_code returns the
    correct string for the STATE property.
    """
    prop = 'STATE'

    result = wypy.translate_status_code(prop, 70)
    assert result == colored('connected', 'green')

    result = wypy.translate_status_code(prop, 60)
    assert result == colored('connected (site)', 'green')

    result = wypy.translate_status_code(prop, 50)
    assert result == colored('connected (local)', 'green')

    result = wypy.translate_status_code(prop, 40)
    assert result == colored('connecting', 'yellow')

    result = wypy.translate_status_code(prop, 30)
    assert result == colored('disconnecting', 'red')

    result = wypy.translate_status_code(prop, 20)
    assert result == colored('disconnected', 'red')

    result = wypy.translate_status_code(prop, 10)
    assert result == colored('asleep', 'yellow')

    result = wypy.translate_status_code(prop, 0)
    assert result == colored('unknown', 'red')




from unittest import mock
from unittest.mock import call
from dbus.proxies import ProxyObject
from wypy.utils.constants import (
    DBUS_GENERAL_PROPS,
    NM_OBJ_PATH,
    NM_IFACE
)


def test_show_status(general, mocker):
    """
    Assert that General.show_status effectively makes calls to
    the following private methods:

        _get_status_info
        translate_status_code

    Both of which are being mocked.
    This test also asserts that translate_status_code was called 6 times.
    """

    prop_names = DBUS_GENERAL_PROPS.values()
    status_codes = [70, 4, 1, 1, 1, 1]
    calls = [
        call(name, status)
        for name, status in zip(prop_names, status_codes)
    ]
    info = [
        {'name': name, 'status_code': status}
        for name, status in zip(prop_names, status_codes)
    ]

    echo_mock = mocker.patch('click.echo')
    mock_info = mocker.patch.object(general, '_get_status_info', return_value=info)
    mock_translate = mocker.patch.object(general, 'translate_status_code')

    general.show_status()

    assert len(echo_mock.call_args_list) == len(info) + 1
    mock_info.assert_called_once()
    mock_translate.assert_has_calls(calls, any_order=True)

def test_get_status_info(general, mocker):
    """
    Assert General.get_status_info calls the get_object_property
    method as many times as there are status_properties,
    each time passing the correct argument.
    """
    mock_get_prop = mocker.patch.object(general, 'get_all_properties')

    general._get_status_info()

    expected_call = call(iface_name=NM_IFACE, object_path=NM_OBJ_PATH)

    assert mock_get_prop.call_args == expected_call



def test_get_hostname(general, mocker):
    """
    Assert General.get_hostname calls the get_object_property method
    with 'Hostname' as the argument.
    """
    dbus_prop_name = 'Hostname'
    mock_hostname = 'some hostname'

    mock_get_prop = mocker.patch.object(general, 'get_object_property', return_value=mock_hostname)
    mock_click_echo = mocker.patch('click.echo')

    general.get_hostname()

    prop_arg = mock_get_prop.call_args[1]['prop_name']

    mock_click_echo.assert_called_once_with(f'Hostname: {mock_hostname}')
    assert prop_arg == dbus_prop_name



import setuptools
from wypy.utils.constants import VERSION

with open("README.md") as f:
    long_description = f.read()

setuptools.setup(
    name='wypy',
    version=VERSION,
    description='A NetworkManager CLI utility.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(
        exclude=['docs', 'tests']
    ),
    url="https://github.com/Zabanaa/wypy",
    author="Karim C",
    author_email="karim.cheurfi@gmail.com",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Topic :: Security'
    ],
    keywords="network-manager network manager nmcli dbus d-bus",
    python_requires='>=3.7',
    install_requires=[
        'click',
        'termcolor',
        'prettytable',
        'dbus-python'
    ],
    entry_points={
        'console_scripts': [
            'wypy=wypy.cli:cli'
        ]
    }
)



from wypy.utils.helpers import is_valid_uuid
from prettytable import PrettyTable
from termcolor import colored
from wypy.utils.constants import (
    NM_IFACE,
    NM_BUS_NAME,
    NM_OBJ_PATH,
    NM_SETTINGS_IFACE,
    NM_SETTINGS_OBJ_PATH,
    NM_ACTIVE_CONN_IFACE,
    NM_DEVICE_IFACE,
    NM_CONNECTION_IFACE
)
from wypy.wypy import WyPy
import click
import dbus
import sys


class Connection(WyPy):

    def __init__(self):
        super().__init__()
        self.proxy = self.bus.get_object(NM_BUS_NAME, NM_OBJ_PATH)
        self.nm_iface = dbus.Interface(self.proxy, NM_IFACE)
        self.active_conns_prop = 'ActiveConnections'
        self.conn_props = ['Id', 'Uuid', 'Type', 'Devices']
        self.table = PrettyTable(['NAME', 'UUID', 'TYPE', 'DEVICE'])
        self.table.sortby = 'TYPE'
        self.table.align = 'l'
        self.active_connections = []
        self.settings_obj = self.bus.get_object(NM_BUS_NAME, NM_SETTINGS_OBJ_PATH)  # noqa E501
        self.settings_iface = dbus.Interface(self.settings_obj, NM_SETTINGS_IFACE)  # noqa E501

    def activate(self):
        click.echo('Activating connection ...')
        """
            TODO:
                - list all connections
                - get the one that matches the given name
                - find out what device it belongs to
                by calling GetDeviceByIpIface()
                - if no connection matches the given name, exit
                with clear and friendly error message.
        """

    def deactivate(self, conn):
        """
        Deactivates a connection.

        Arguments:
            conn {string} -- uuid / name of the connection to delete
        """
        click.echo('Deactivating connection ...')
        self._get_active_connections()
        filter_key = 'Uuid' if is_valid_uuid(conn) else 'Id'
        conn_ids = [str(c[filter_key]) for c in self.active_connections]

        if conn not in conn_ids:
            sys.exit(f'Could not deactivate {conn}. Connection unknown or inactive.')  # noqa E501
        else:
            conn_to_deactivate = next(filter(lambda x: str(x[filter_key]) == conn, self.active_connections), None)  # noqa E501
            self._deactivate_connection(conn_to_deactivate['Path'])

    def delete(self, conn):
        """
        Deletes a connection profile.
        If conn is a valid uuid, it will be used to retrieve the connection.
        If not, WyPy will try to filter all known connections
        until it finds a match.
        If no match is found, the program exits with an error message.

        Arguments:
            conn {[string]} -- [uuid / name of the connection to delete]
        """
        click.echo(f'Deleting connection profile {conn}')
        if is_valid_uuid(conn):
            uuid = conn
        else:
            connections = self._list_connections_info()
            conn_to_delete = next(filter(lambda x: str(x['id']) == conn, connections), None)  # noqa E501
            try:
                uuid = conn_to_delete['uuid']
            except TypeError:
                sys.exit(f'Could not delete {conn}. Connection unknown or already deleted.')  # noqa E501

        self._delete_connection(uuid)

    def show_all(self):
        """
        Echos the list of all available connections to the user.
        """
        click.echo('Showing all connections ...')
        self._get_active_connections()
        connections = self._list_connections_info()
        rows = list(map(self._create_row, connections))

        for row in rows:
            self.table.add_row(row)

        click.echo(self.table)

    def show_active(self):
        """
        Echos the list of active connections to the user.
        """
        click.echo('Showing active connections ...')
        self._get_active_connections()
        conn_values = [conn.values() for conn in self.active_connections]

        for i, vals in enumerate(conn_values):
            conn_values[i] = list(map(lambda val: colored(val, "green"), vals))

        self._print_conns(conn_values)

    #   ---------------
    #
    #   Private Methods
    #
    #   ---------------

    def _delete_connection(self, uuid):
        """
        Gets the connection object to delete by uuid
        Creates the associated interface
        Calls the Delete method on that interface

        Arguments:
            uuid {string} -- the uuid of the connection to delete
        """
        _conn = self.settings_iface.GetConnectionByUuid(uuid)
        conn_obj = self.bus.get_object(NM_BUS_NAME, _conn)
        conn_iface = dbus.Interface(conn_obj, NM_CONNECTION_IFACE)
        conn_iface.Delete()

    def _deactivate_connection(self, conn_path):
        """
        Calls DeactivateConnection on the main d-bus
        interface provided by NetworkManager.

        Arguments:
            conn_path {string} -- the connection's own d-bus object path
        """
        self.nm_iface.DeactivateConnection(conn_path)

    def _create_row(self, conn):

        conn_info = {
            'id': str(conn.get('id', '')),
            'uuid': str(conn.get('uuid', '')),
            'type': conn.get('type', ''),
            'name': str(conn.get('interface-name', '--')),
        }
        values = conn_info.values()

        if self._is_connection_active(conn_info['uuid']):
            values = list(map(lambda val: colored(val, "green"), values))

        return values

    def _get_active_connections(self):
        """
        Retrieves the list of active connection on the system.
        For each active connection found, a dict containing only
        the necessary information is created and appended to
        self.active_connections.
        """
        active_conns_paths = self.get_object_property(
            proxy=self.proxy,
            prop_name=self.active_conns_prop
        )

        for conn in active_conns_paths:
            all_props = self.get_all_properties(conn, NM_ACTIVE_CONN_IFACE)
            conn_props = {
                prop: str(all_props[prop])
                for prop in self.conn_props
                if not isinstance(all_props[prop], dbus.Array)
            }

            device_name, device_type = self._get_device_info(all_props)
            conn_props['Device'] = device_name
            conn_props['Type'] = self.translate_device_type(device_type)
            conn_props['Path'] = conn

            self.active_connections.append(conn_props)

    def _get_device_info(self, connection_props):
        """
        Extract both the name and type of the device associated with
        the given connection.

        Arguments:
            connection_props {[dict]} -- [the connection's properties]

        Returns:
            [tuple] -- [device name and type associated with the connection]
        """
        device_obj = str(connection_props['Devices'][0])
        device_props = self.get_all_properties(device_obj, NM_DEVICE_IFACE)
        return (device_props['Interface'], int(device_props['DeviceType']))

    def _print_conns(self, data):
        """
        Populates self.table with the connection information.
        Displays the table to the user.

        Arguments:
            data {[list]} -- [the list of available connections]
        """
        for row in data:
            self.table.add_row(row)
        click.echo(self.table)

    def _is_connection_active(self, uuid):
        """
        Tests whether or not the given connection is active.
        Arguments:
            uuid {[string]} -- [the uuid of the connection to test]

        Returns:
            [bool] -- [whether the connection is active or not]
        """
        active_conn_uuids = [conn['Uuid'] for conn in self.active_connections]
        return uuid in active_conn_uuids

    def _color_data(self, conn):
        """
        Colors the values of the given connection dictionary.

        Arguments:
            conn {[dict]} -- [the connection whose values are to be colored]
        """
        for key, val in conn.items():
            conn[key] = colored(val, "green")

    def _list_connections_info(self):
        """
        Calls ListConnections available on the settings
        d-bus interface provided by NetworkManager.
        For each connection path, an connection interface is
        created and the GetSettings method is called.
        The 'connection' property is then accessed and assigned to
        a dictionary, which is itself appended to the result list.

        Returns:
            [list] -- [dictionaries containing information
            about the available connections.]
        """
        connections = self.settings_iface.ListConnections()
        result = []

        for conn in connections:
            conn_obj = self.bus.get_object(NM_BUS_NAME, conn)
            conn_iface = dbus.Interface(conn_obj,  NM_CONNECTION_IFACE)
            conn_info = conn_iface.GetSettings()['connection']
            result.append(conn_info)

        return result



import click
from .general import General


@click.group('general')
@click.pass_context
def general(ctx):
    """Gather general system information"""
    ctx.obj = General()


@click.command()
@click.pass_obj
def status(gen):
    """Print general NetworkManager status"""
    gen.show_status()


@general.command()
@click.pass_obj
def hostname(gen):
    """Print hostname"""
    gen.get_hostname()


general.add_command(status)
general.add_command(hostname)



import click
import sys
from termcolor import colored
from wypy.utils.helpers import nm_is_installed
from wypy.device.commands import device as device_commands
from wypy.general.commands import general as general_commands
from wypy.networking.commands import network as network_commands
from wypy.wifi.commands import wifi as wifi_commands
from wypy.connection.commands import connection as conn_commands
from wypy.utils.constants import VERSION


@click.group()
@click.version_option(version=VERSION, message=f'WyPy - Version {VERSION}')
def cli():
    """
    wypy is a command line utility for NetworkManager.
    """
    pass


cli.add_command(general_commands)
cli.add_command(network_commands)
cli.add_command(wifi_commands)
cli.add_command(conn_commands)
cli.add_command(device_commands)

if __name__ == "__main__":
    is_nm_installed = nm_is_installed()
    if is_nm_installed:
        from dbus.mainloop.glib import DBusGMainLoop
        DBusGMainLoop(set_as_default=True)
        cli()
    else:
        msg = """
        [Error]: NetworkManager must be installed on the system to use WyPy.
        """.replace('  ', '')
        sys.exit(colored(msg, "red"))



from wypy.wypy import WyPy
from wypy.general import General
from wypy.device import Device
from wypy.wifi import WiFi
from wypy.networking import Network
import pytest


@pytest.fixture(scope='function')
def general():
    general = General()
    yield general


@pytest.fixture(scope='function')
def wypy():
    wypy = WyPy()
    yield wypy


@pytest.fixture(scope='function')
def device():
    device = Device()
    yield device


@pytest.fixture(scope='function')
def network():
    network = Network()
    yield network


@pytest.fixture(scope='function')
def wifi():
    wifi = WiFi()
    yield wifi



import click
from .device import Device


@click.group('device')
@click.pass_context
def device(ctx):
    """Perform actions on the available device(s)"""
    ctx.obj = Device()


@click.command('status')
@click.pass_obj
def status(dev):
    """Print general device status information"""
    dev.print_status()


@click.command('list')
@click.pass_obj
def list_all(dev):
    """List detailed device information"""
    dev.list_all()


@click.command('get')
@click.argument('name', type=str, required=True)
@click.pass_obj
def get(dev, name):
    """List detailed device information for a given device"""
    dev.print_details(name)


@click.command('update')
@click.argument('ifname', type=str, required=True)
@click.pass_obj
def update(dev, ifname):
    """
    Update device by applying the last changes done
    to its active connection
    """
    dev.update_ifname_connection(ifname)


@click.command('disconnect')
@click.argument('ifname', type=str, required=True)
@click.pass_obj
def disconnect(dev, ifname):
    """Disconnect the device"""
    dev.disconnect(ifname)


@click.command('delete')
@click.argument('ifname', type=str, required=True)
@click.pass_obj
def delete(dev, ifname):
    """Delete the device"""
    dev.delete_iface(ifname)


@click.command('manage')
@click.option(
    '--off',
    is_flag=True,
    help='Disable device management by NetworkManager'
)
@click.argument('ifname', type=str, required=True)
@click.pass_obj
def manage(dev, ifname, off):
    """Manage / Unmanage the device"""
    if off:
        dev.manage(ifname, flag=False)
    else:
        dev.manage(ifname)


@click.command('autoconnect')
@click.option('--disable', is_flag=True, help='Disable autoconnect')
@click.argument('ifname', type=str, required=True)
@click.pass_obj
def autoconnect(dev, ifname, disable):
    """Enable / Disable autoconnect on the device"""
    if disable:
        dev.autoconnect(ifname, flag=False)
    else:
        dev.autoconnect(ifname)


device.add_command(status)
device.add_command(list_all)
device.add_command(get)
device.add_command(update)
device.add_command(disconnect)
device.add_command(delete)
device.add_command(manage)
device.add_command(autoconnect)



import click
from .networking import Network


@click.group('network')
@click.pass_context
def network(ctx):
    """Perform networking actions"""
    ctx.obj = Network()


@click.command('on')
@click.pass_obj
def turn_on(network):
    """Turn networking capabiliy on"""
    network.turn_on()


@click.command('off')
@click.pass_obj
def turn_off(network):
    """Turn networking capabiliy off"""
    network = network.turn_off()


@click.command('connectivity')
@click.option(
    '--check',
    is_flag=True,
    help='force WyPy to recheck connectivity state'
)
@click.pass_obj
def connectivity(network, check):
    """
    Get connectivity state.
    If check is passed, WyPy will perform a check on NetworkManager
    """
    if check:
        network.check_connectivity()
    else:
        network.get_connectivity_state()


network.add_command(turn_on)
network.add_command(turn_off)
network.add_command(connectivity)
